// Copyright 2022 The IREE Authors
//
// Licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

#include "CommonExtensions.h"

#include "iree-dialects/Dialect/LinalgTransform/StructuredTransformOpsExt.h"
#include "iree-dialects/Transforms/ListenerGreedyPatternRewriteDriver.h"
#include "iree/compiler/Codegen/Interfaces/BufferizationInterfaces.h"
#include "iree/compiler/Codegen/Passes.h"
#include "llvm/ADT/StringSet.h"
#include "mlir/Pass/PassManager.h"

using namespace mlir;
using namespace mlir::iree_compiler;
using namespace mlir::iree_compiler::IREE;

iree_compiler::IREE::transform_dialect::CommonExtensions::CommonExtensions() {
  registerTransformOps<
#define GET_OP_LIST
#include "iree/compiler/Codegen/Common/TransformExtensions/CommonExtensionsOps.cpp.inc"
      >();
}

void mlir::iree_compiler::registerTransformDialectCommonExtension(
    DialectRegistry &registry) {
  registry.addExtensions<transform_dialect::CommonExtensions>();
}

//===---------------------------------------------------------------------===//
// ApplyPatternsOp
//===---------------------------------------------------------------------===//

static void addRankReducingPatterns(RewritePatternSet &patterns) {
  vector::populateCastAwayVectorLeadingOneDimPatterns(patterns);
  linalg::populateFoldUnitExtentDimsPatterns(patterns);
}

static void addAllRegisteredCanonicalizationPatterns(
    RewritePatternSet &patterns) {
  MLIRContext *ctx = patterns.getContext();
  for (Dialect *dialect : ctx->getLoadedDialects())
    dialect->getCanonicalizationPatterns(patterns);
  for (RegisteredOperationName op : ctx->getRegisteredOperations())
    op.getCanonicalizationPatterns(patterns, ctx);
}

DiagnosedSilenceableFailure transform_dialect::ApplyPatternsOp::applyToOne(
    Operation *target, SmallVectorImpl<Operation *> &results,
    transform::TransformState &state) {
  if (!target->hasTrait<OpTrait::IsIsolatedFromAbove>()) {
    target->emitOpError(
        "applies only to isolated-from-above targets because it needs to apply "
        "patterns greedily");
    return DiagnosedSilenceableFailure(reportUnknownTransformError(target));
  }
  MLIRContext *ctx = target->getContext();
  RewritePatternSet patterns(ctx);
  if (getCanonicalization()) addAllRegisteredCanonicalizationPatterns(patterns);
  if (getRankReducing()) addRankReducingPatterns(patterns);

  TrackingListener listener(state);
  GreedyRewriteConfig config;
  LogicalResult result = applyPatternsAndFoldGreedily(
      target, std::move(patterns), config, &listener);
  LogicalResult listenerResult = listener.checkErrorState();
  if (failed(result) || failed(listenerResult))
    return DiagnosedSilenceableFailure(reportUnknownTransformError(target));
  results.assign({target});
  return DiagnosedSilenceableFailure(success());
}

//===---------------------------------------------------------------------===//
// IREEBufferizeOp
//===---------------------------------------------------------------------===//

//===---------------------------------------------------------------------===//
// Default allocation functions for CPU backend
// TODO: register the bufferization behavior in a target-specific way.
// TODO: Maybe bufferize should have a separate cpu and a gpu version. This is
// unclear though: what happens on heterogeneous HW ?
//===---------------------------------------------------------------------===//

// Allocation callbacks to use with upstream comprehensive bufferization
static FailureOr<Value> cpuComprehensiveBufferizeAllocationFn(
    OpBuilder &builder, Location loc, MemRefType memRefType,
    ValueRange dynamicSizes, unsigned alignment) {
  return builder
      .create<memref::AllocaOp>(loc, memRefType, dynamicSizes,
                                builder.getI64IntegerAttr(alignment))
      .getResult();
}

static LogicalResult cpuComprehensiveBufferizeDeallocationFn(OpBuilder &builder,
                                                             Location loc,
                                                             Value allocation) {
  return success();
}

static LogicalResult cpuComprehensiveBufferizeCopyFn(OpBuilder &builder,
                                                     Location loc, Value from,
                                                     Value to) {
  // TODO: ideally we should use linalg.copy which was recently reintroduced
  // as an OpDSL named op. However, IREE-specific patterns to cleanup spurious
  // post-bufferization copies do not trigger properly.
  // So we keep using `createLinalgCopyOp` which builds a GenericOp.
  // builder.create<linalg::CopyOp>(loc, from, to);
  mlir::iree_compiler::createLinalgCopyOp(builder, loc, from, to);
  return success();
}

static FailureOr<Value> gpuComprehensiveBufferizeAllocationFn(
    OpBuilder &builder, Location loc, MemRefType memRefType,
    ValueRange dynamicSizes, unsigned alignment) {
  // TODO: use gpu::GPUDialect::getWorkgroupAddressSpace() but this requires
  // moving out of CommonExtensions.
  MemRefType allocType = MemRefType::get(memRefType.getShape(),
                                         memRefType.getElementType(), {}, 3);
  return builder
      .create<memref::AllocOp>(loc, allocType, dynamicSizes,
                               builder.getI64IntegerAttr(alignment))
      .getResult();
}

static LogicalResult gpuComprehensiveBufferizeDeallocationFn(OpBuilder &builder,
                                                             Location loc,
                                                             Value allocation) {
  builder.create<memref::DeallocOp>(loc, allocation);
  return success();
}

static LogicalResult gpuComprehensiveBufferizeCopyFn(OpBuilder &builder,
                                                     Location loc, Value from,
                                                     Value to) {
  // TODO: ideally we should use linalg.copy which was recently reintroduced
  // as an OpDSL named op. However, IREE-specific patterns to cleanup spurious
  // post-bufferization copies do not trigger properly.
  // So we keep using `createLinalgCopyOp` which builds a GenericOp.
  // builder.create<linalg::CopyOp>(loc, from, to);
  mlir::iree_compiler::createLinalgCopyOp(builder, loc, from, to);
  return success();
}

DiagnosedSilenceableFailure transform_dialect::IREEBufferizeOp::apply(
    transform::TransformResults &results, transform::TransformState &state) {
  PassManager pm(getContext());
  // Bufferize the dispatch.
  using mlir::bufferization::BufferizationOptions;
  BufferizationOptions::AllocationFn allocationFn =
      cpuComprehensiveBufferizeAllocationFn;
  BufferizationOptions::DeallocationFn deallocationFn =
      cpuComprehensiveBufferizeDeallocationFn;
  BufferizationOptions::MemCpyFn memcpyFn = cpuComprehensiveBufferizeCopyFn;
  if (getTargetGpu()) {
    allocationFn = gpuComprehensiveBufferizeAllocationFn;
    deallocationFn = gpuComprehensiveBufferizeDeallocationFn;
    memcpyFn = gpuComprehensiveBufferizeCopyFn;
  }
  mlir::iree_compiler::addIREEComprehensiveBufferizePasses(
      pm, allocationFn, deallocationFn, memcpyFn);
  WalkResult res = state.getTopLevel()->walk([&](ModuleOp moduleOp) {
    if (failed(pm.run(moduleOp))) {
      getOperation()->emitError()
          << "failed to bufferize ModuleOp:\n"
          << *(moduleOp.getOperation()) << "\nunder top-level:\n"
          << *state.getTopLevel();
      return WalkResult::interrupt();
    }
    return WalkResult::advance();
  });
  return DiagnosedSilenceableFailure(failure(res.wasInterrupted()));
}

//===---------------------------------------------------------------------===//
// MatchOp
//===---------------------------------------------------------------------===//

LogicalResult transform_dialect::MatchOp::verify() {
  bool opXorIface = getMatchOp().hasValue() ^ getMatchInterface().hasValue();
  if (!opXorIface)
    return this->emitOpError(
        "requires a either a match_op or a match_interface attribute (but not "
        "both)");
  return success();
}

DiagnosedSilenceableFailure transform_dialect::MatchOp::apply(
    transform::TransformResults &results, transform::TransformState &state) {
  llvm::StringSet<> strs;
  if (getMatchOp().hasValue())
    strs.insert(getMatchOp()->getAsValueRange<StringAttr>().begin(),
                getMatchOp()->getAsValueRange<StringAttr>().end());

  ArrayRef<Operation *> payloadOps = state.getPayloadOps(getTarget());
  if (payloadOps.size() != 1)
    return DiagnosedSilenceableFailure(
        this->emitOpError("requires exactly one target handle"));

  SmallVector<Operation *> res;

  auto matchFun = [&](Operation *op) {
    if (strs.contains(op->getName().getStringRef())) res.push_back(op);
    // Interfaces cannot be matched by name, just by ID.
    // So we specifically encode the interfaces we care about for this op.
    if (getMatchInterface().hasValue()) {
      auto iface = getMatchInterface().getValue();
      if (iface == transform_dialect::MatchInterfaceEnum::LinalgOp &&
          isa<linalg::LinalgOp>(op))
        res.push_back(op);
      if (iface == transform_dialect::MatchInterfaceEnum::TilingInterface &&
          isa<TilingInterface>(op))
        res.push_back(op);
    }
  };

  payloadOps.front()->walk(matchFun);
  results.set(getResult().cast<OpResult>(), res);
  return DiagnosedSilenceableFailure(success());
}

ParseResult transform_dialect::MatchOp::parse(OpAsmParser &parser,
                                              OperationState &result) {
  // Parse 'match_op' or 'interface' clause.
  if (succeeded(parser.parseOptionalKeyword("ops"))) {
    ArrayAttr opsAttr;
    if (parser.parseLBrace() ||
        parser.parseCustomAttributeWithFallback(
            opsAttr, parser.getBuilder().getType<NoneType>(), "match_op",
            result.attributes) ||
        parser.parseRBrace())
      return failure();
  } else if (succeeded(parser.parseOptionalKeyword("interface"))) {
    if (parser.parseLBrace()) return failure();
    StringRef attrStr;
    auto loc = parser.getCurrentLocation();
    if (parser.parseKeyword(&attrStr)) return failure();
    auto interfaceEnum =
        transform_dialect::symbolizeMatchInterfaceEnum(attrStr);
    if (!interfaceEnum)
      return parser.emitError(loc, "invalid ")
             << "match_interface attribute specification: \"" << attrStr << '"';
    transform_dialect::MatchInterfaceEnumAttr match_interfaceAttr =
        transform_dialect::MatchInterfaceEnumAttr::get(
            parser.getBuilder().getContext(), interfaceEnum.value());
    result.addAttribute("match_interface", match_interfaceAttr);
    if (parser.parseRBrace()) return failure();
  } else {
    auto loc = parser.getCurrentLocation();
    return parser.emitError(loc, "expected ops or interface");
  }

  OpAsmParser::UnresolvedOperand targetRawOperands[1];
  ArrayRef<OpAsmParser::UnresolvedOperand> targetOperands(targetRawOperands);
  if (parser.parseKeyword("in") || parser.parseOperand(targetRawOperands[0]) ||
      parser.parseOptionalAttrDict(result.attributes))
    return failure();
  Type pdlOpType = parser.getBuilder().getType<pdl::OperationType>();
  result.addTypes(pdlOpType);
  if (parser.resolveOperands(targetOperands, pdlOpType, result.operands))
    return failure();
  return success();
}

void transform_dialect::MatchOp::print(OpAsmPrinter &p) {
  if ((*this)->getAttr("match_op")) {
    p << " ops{";
    p.printAttributeWithoutType(getMatchOpAttr());
    p << "}";
  }
  if ((*this)->getAttr("match_interface")) {
    p << " interface{" << stringifyMatchInterfaceEnum(*getMatchInterface())
      << "}";
  }
  p << " in " << getTarget();
  p.printOptionalAttrDict((*this)->getAttrs(),
                          /*elidedAttrs=*/{"match_op", "match_interface"});
}

#include "iree/compiler/Codegen/Common/TransformExtensions/CommonExtensionsAttrs.cpp.inc"

#define GET_OP_CLASSES
#include "iree/compiler/Codegen/Common/TransformExtensions/CommonExtensionsOps.cpp.inc"
