// Copyright 2020 The IREE Authors
//
// Licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

#include <utility>

#include "iree/compiler/Dialect/Flow/IR/FlowDialect.h"
#include "iree/compiler/Dialect/Flow/IR/FlowOps.h"
#include "iree/compiler/Dialect/Flow/Transforms/PassDetail.h"
#include "iree/compiler/Dialect/Flow/Transforms/Passes.h"
#include "iree/compiler/Dialect/Util/IR/UtilDialect.h"
#include "iree/compiler/Dialect/Util/IR/UtilOps.h"
#include "mlir/IR/Attributes.h"
#include "mlir/IR/Builders.h"
#include "mlir/Pass/Pass.h"

namespace mlir {
namespace iree_compiler {
namespace IREE {
namespace Flow {

class StripAndSplatConstantVariablesPass
    : public StripAndSplatConstantVariablesBase<
          StripAndSplatConstantVariablesPass> {
 public:
  StripAndSplatConstantVariablesPass() = default;

  void getDependentDialects(DialectRegistry &registry) const override {
    registry.insert<IREE::Flow::FlowDialect, IREE::Util::UtilDialect>();
  }

  void runOnOperation() override {
    auto moduleOp = getOperation();
    auto builder = OpBuilder::atBlockBegin(moduleOp.getBody());

    // Use a heuristic to space out splat values in hopes of avoiding NaN and
    // INF values at runtime:
    //   floats: 1/1, 1/2, 1/3, ...
    //   ints: 1, 2, 3, 4, ...
    // TODO(scotttodd): flags to control numbers used (all 0, all 1, increasing)
    int replaceIndex = 1;

    moduleOp.walk([&](IREE::Util::GlobalOp op) {
      // Only strip constant variables.
      if (op.getIsMutable()) {
        return;
      }

      // Only strip tensor type constants (to replace with dense<>).
      if (!op.getType().isa<TensorType>()) {
        return;
      }

      auto tensorType = op.getType().cast<TensorType>();
      auto elementType = tensorType.getElementType();
      DenseElementsAttr newValue;
      if (elementType.isa<FloatType>()) {
        newValue = DenseElementsAttr::get(
            tensorType, FloatAttr::get(elementType, 1.0 / replaceIndex));
      } else {
        newValue = DenseElementsAttr::get(
            tensorType, IntegerAttr::get(elementType, replaceIndex));
      }

      builder.setInsertionPointAfter(op);
      auto newOp = builder.create<IREE::Util::GlobalOp>(
          op.getLoc(), op.getSymName(), op.getIsMutable(), op.getType(),
          newValue);
      newOp.setVisibility(op.getVisibility());
      newOp->setAttr("noinline", UnitAttr::get(builder.getContext()));
      op.erase();

      replaceIndex++;
    });
  }
};

std::unique_ptr<OperationPass<mlir::ModuleOp>>
createStripAndSplatConstantVariablesPass() {
  return std::make_unique<StripAndSplatConstantVariablesPass>();
}

}  // namespace Flow
}  // namespace IREE
}  // namespace iree_compiler
}  // namespace mlir
